"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const axios_1 = require("axios");
const cheerio = require("cheerio");
const url_1 = require("url");
const vscode_1 = require("vscode");
const fs = require("fs");
const path_1 = require("path");
function mkCommandUri(commandName, ...args) {
    return `command:${commandName}?${encodeURIComponent(JSON.stringify(args))}`;
}
exports.mkCommandUri = mkCommandUri;
function findProjectDocumentation() {
    const html = path_1.join(vscode_1.workspace.rootPath, 'html', 'index.html');
    return fs.existsSync(html) ? html : null;
}
class DocViewProvider {
    constructor(staticServer) {
        this.staticServer = staticServer;
        this.subscriptions = [];
        this.currentURL = undefined;
        this.backstack = [];
        this.forwardstack = [];
        this.subscriptions.push(vscode_1.commands.registerCommand('lean.openDocView', (url) => this.open(url)), vscode_1.commands.registerCommand('lean.backDocView', () => this.back()), vscode_1.commands.registerCommand('lean.forwardDocView', () => this.forward()), vscode_1.commands.registerCommand('lean.openTryIt', (code) => this.tryIt(code)), vscode_1.commands.registerCommand('lean.openExample', (file) => this.example(file)));
        this.offerToOpenProjectDocumentation();
    }
    offerToOpenProjectDocumentation() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!fs.existsSync(path_1.join(vscode_1.workspace.rootPath, 'leanpkg.toml')))
                return;
            const projDoc = findProjectDocumentation();
            if (!projDoc)
                return;
            yield this.open(vscode_1.Uri.file(projDoc).toString());
        });
    }
    tryIt(code) {
        return __awaiter(this, void 0, void 0, function* () {
            const doc = yield vscode_1.workspace.openTextDocument({ language: 'lean', content: code });
            const editor = yield vscode_1.window.showTextDocument(doc, vscode_1.ViewColumn.One);
        });
    }
    example(file) {
        return __awaiter(this, void 0, void 0, function* () {
            const doc = yield vscode_1.workspace.openTextDocument(vscode_1.Uri.parse(file));
            yield vscode_1.window.showTextDocument(doc, vscode_1.ViewColumn.One);
        });
    }
    getWebview() {
        if (!this.webview) {
            const options = {
                enableFindWidget: true,
                enableScripts: true,
                enableCommandUris: true,
                retainContextWhenHidden: true,
            };
            this.webview = vscode_1.window.createWebviewPanel('lean', 'Lean Documentation', { viewColumn: 3, preserveFocus: true }, options);
            this.webview.onDidDispose(() => this.webview = null);
        }
        return this.webview;
    }
    fetch(url) {
        return __awaiter(this, void 0, void 0, function* () {
            if (url) {
                const uri = vscode_1.Uri.parse(url);
                if (uri.scheme === 'file') {
                    if (uri.fsPath.endsWith('.html')) {
                        return fs.readFileSync(uri.fsPath).toString();
                    }
                }
                else {
                    const { data, headers } = yield axios_1.default.get(url);
                    if (('' + headers['content-type']).startsWith('text/html')) {
                        return data;
                    }
                }
                const $ = cheerio.load('<p>');
                $('p').text('Unsupported file. ')
                    .append($('<a>').attr('href', url).attr('alwaysExternal', 'true')
                    .text('Open in browser instead.'));
                return $.html();
            }
            else {
                const $ = cheerio.load('<body>');
                const body = $('body');
                const html = findProjectDocumentation();
                if (html) {
                    body.append($('<p>').append($('<a>').attr('href', vscode_1.Uri.file(html).toString())
                        .text('Open documentation of current project')));
                }
                const books = {
                    'Theorem Proving in Lean': 'https://leanprover.github.io/theorem_proving_in_lean/',
                    'Reference Manual': 'https://leanprover.github.io/reference/',
                    'Mathematics in Lean': 'https://avigad.github.io/mathematics_in_lean/',
                };
                for (const book of Object.getOwnPropertyNames(books)) {
                    body.append($('<p>').append($('<a>').attr('href', books[book]).text(book)));
                }
                return $.html();
            }
        });
    }
    mkRelativeUrl(relative, base) {
        const uri = new url_1.URL(relative, base);
        if (uri.protocol === 'file:') {
            if (new url_1.URL(base).protocol !== 'file:') {
                return '';
            }
            const path = vscode_1.Uri.parse(uri.toString()).fsPath;
            // workaround for https://github.com/microsoft/vscode/issues/89038
            return this.staticServer.mkUri(path);
        }
        else {
            return uri.toString();
        }
    }
    setHtml() {
        return __awaiter(this, void 0, void 0, function* () {
            const { webview } = this.getWebview();
            const url = this.currentURL;
            let $;
            try {
                $ = cheerio.load(yield this.fetch(url));
            }
            catch (e) {
                $ = cheerio.load('<pre>');
                $('pre').text(e.toString());
            }
            for (const style of $('link[rel=stylesheet]').get()) {
                style.attribs.href = this.mkRelativeUrl(style.attribs.href, url);
            }
            for (const script of $('script[src]').get()) {
                script.attribs.src = this.mkRelativeUrl(script.attribs.src, url);
            }
            for (const link of $('a[href]').get()) {
                const tryItMatch = link.attribs.href.match(/\/(?:live|lean-web-editor)\/.*#code=(.*)/);
                if (link.attribs.href.startsWith('#')) {
                    // keep relative links
                }
                else if (link.attribs.alwaysExternal) {
                    // keep links with alwaysExternal attribute
                }
                else if (link.attribs.tryitfile) {
                    link.attribs.title = link.attribs.title || 'Open code block (in existing file)';
                    link.attribs.href = mkCommandUri('lean.openExample', new url_1.URL(link.attribs.tryitfile, url).toString());
                }
                else if (tryItMatch) {
                    const code = decodeURIComponent(tryItMatch[1]);
                    link.attribs.title = link.attribs.title || 'Open code block in new editor';
                    link.attribs.href = mkCommandUri('lean.openTryIt', code);
                }
                else {
                    const hrefUrl = new url_1.URL(link.attribs.href, url);
                    const isExternal = url && new url_1.URL(url).origin !== hrefUrl.origin;
                    if (!isExternal) {
                        link.attribs.title = link.attribs.title || link.attribs.href;
                        link.attribs.href = mkCommandUri('lean.openDocView', hrefUrl.toString());
                    }
                }
            }
            // javascript search doesn't work
            $('.sphinxsidebar #searchbox').remove();
            const nav = $('<nav>');
            nav.css('width', '100vw');
            nav.css('position', 'fixed');
            nav.css('top', '0');
            nav.css('right', '0');
            nav.css('text-align', 'right');
            nav.css('z-index', '100');
            nav.prependTo($('body'));
            const navDiv = $('<span>');
            navDiv.css('padding', '4px');
            navDiv.css('padding-right', '20px');
            navDiv.css('z-index', '100');
            navDiv.css('background-color', 'var(--vscode-tab-activeBackground)');
            nav.append(navDiv);
            const fontSize = vscode_1.workspace.getConfiguration('editor').get('fontSize') + 'px';
            const mkLink = (command, title, text) => $('<a>')
                .attr('title', title)
                .attr('href', mkCommandUri(command))
                .css('color', 'var(--vscode-tab-activeForeground)')
                .css('font-family', 'sans-serif')
                .css('font-size', fontSize)
                .css('margin-left', '1em')
                .css('text-decoration', 'none')
                .text(text);
            navDiv.append(mkLink('lean.backDocView', 'back', '⬅ back'));
            navDiv.append(mkLink('lean.forwardDocView', 'forward', 'forward ➡'));
            $('nav+*').css('margin-top', '3em');
            webview.html = $.html();
        });
    }
    /** Called by the user clicking a link. */
    open(url) {
        return __awaiter(this, void 0, void 0, function* () {
            if (url) {
                this.backstack.push(this.currentURL);
                this.forwardstack = [];
            }
            this.currentURL = url;
            yield this.setHtml();
        });
    }
    back() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.backstack.length === 0) {
                return;
            }
            this.forwardstack.push(this.currentURL);
            this.currentURL = this.backstack.pop();
            yield this.setHtml();
        });
    }
    forward() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.forwardstack.length === 0) {
                return;
            }
            this.backstack.push(this.currentURL);
            this.currentURL = this.forwardstack.pop();
            yield this.setHtml();
        });
    }
    dispose() {
        for (const s of this.subscriptions) {
            s.dispose();
        }
    }
}
exports.DocViewProvider = DocViewProvider;
//# sourceMappingURL=docview.js.map