"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const vscode_1 = require("vscode");
/** Pastes suggestions provided by tactics such as `squeeze_simp` */
class TacticSuggestions {
    constructor(server, infoView, leanDocs) {
        this.server = server;
        this.leanDocs = leanDocs;
        this.subscriptions = [];
        // Match everything after "Try this" until the next unindented line
        this.magicWord = 'Try this: ';
        this.regex = '^' + this.magicWord + '((.*\n )*.*)$';
        this.regexGM = new RegExp(this.regex, 'gm');
        this.regexM = new RegExp(this.regex, 'm');
        const commandHandler = (textEditor) => {
            const msg = this.findSelectedMessage(textEditor);
            if (msg === null)
                return;
            this.pasteIntoEditor(msg, textEditor, null);
        };
        const infoViewCommandHandler = (m, suggestion) => {
            const textEditor = this.findTextEditor(m.file_name);
            this.pasteIntoEditor(m, textEditor, suggestion);
            // Focus text editor
            vscode_1.window.showTextDocument(textEditor.document, { viewColumn: textEditor.viewColumn });
        };
        this.subscriptions.push(vscode_1.commands.registerTextEditorCommand('lean.pasteTacticSuggestion', commandHandler), vscode_1.commands.registerCommand('_lean.pasteTacticSuggestion', infoViewCommandHandler), vscode_1.languages.registerCodeActionsProvider(this.leanDocs, this));
        infoView.addMessageFormatter((text, m) => {
            const newText = text.replace(this.regexGM, (_, tactic) => {
                const command = encodeURI('command:_lean.pasteTacticSuggestion?' +
                    JSON.stringify([m, tactic]));
                return `${this.magicWord}<a href="${command}" title="${tactic}">${tactic}</a>`;
            });
            return newText;
        });
    }
    dispose() {
        for (const s of this.subscriptions) {
            s.dispose();
        }
    }
    findTextEditor(fileName) {
        for (const textEditor of vscode_1.window.visibleTextEditors) {
            if (textEditor.document.uri.toString() === vscode_1.Uri.file(fileName).toString()) {
                return textEditor;
            }
        }
    }
    findSelectedMessage(textEditor) {
        const curFileName = textEditor.document.fileName;
        const curPosition = textEditor.selection.active;
        // Find message closest to the cursor
        const messages = this.server.messages
            .filter((m) => m.file_name === curFileName &&
            m.pos_line === curPosition.line + 1 &&
            m.pos_col <= curPosition.character)
            .sort((a, b) => b.pos_col - a.pos_col);
        if (messages.length === 0)
            return null;
        return messages[0];
    }
    pasteIntoEditor(m, textEditor, suggestion) {
        return __awaiter(this, void 0, void 0, function* () {
            if (suggestion === null) {
                // Find first suggestion in message
                const suggs = this.regexM.exec(m.text);
                if (!suggs)
                    return;
                suggestion = suggs[1];
            }
            // Start of the tactic call to replace
            const startLine = m.pos_line - 1;
            const startCol = m.pos_col;
            // Try to determine the end of the tactic call to replace.
            // Heuristic: Find the next comma, semicolon, unmatched closing bracket,
            // or newline that is not enclosed in brackets. To keep things simple,
            // we use only one counter for all kinds of brackets.
            let openBrackets = 0;
            let endLine;
            let endCol;
            lineLoop: for (endLine = startLine; endLine < textEditor.document.lineCount; endLine++) {
                const chars = textEditor.document.lineAt(endLine).text.split('').entries();
                // Iterate over every character of the line
                for (const [col, char] of chars) {
                    // Only search from where the tactic starts
                    if (endLine > startLine || col > startCol) {
                        if (openBrackets === 0 && [',', ';'].includes(char)) {
                            endCol = col;
                            break lineLoop;
                        }
                        if (['(', '[', '{', '⟨', '⦃'].includes(char)) {
                            openBrackets++;
                        }
                        if (['⦄', '⟩', '}', ']', ')'].includes(char)) {
                            if (openBrackets === 0) {
                                endCol = col;
                                break lineLoop;
                            }
                            else {
                                openBrackets--;
                            }
                        }
                    }
                }
                if (openBrackets === 0) {
                    endCol = textEditor.document.lineAt(endLine).range.end.character;
                    break lineLoop;
                }
            }
            // Jump to the end of the tactic call
            const lastPos = new vscode_1.Position(endLine, endCol);
            textEditor.selection = new vscode_1.Selection(lastPos, lastPos);
            // Replace tactic call by suggestion
            const range = new vscode_1.Range(new vscode_1.Position(startLine, startCol), new vscode_1.Position(endLine, endCol));
            yield textEditor.edit(editBuilder => {
                editBuilder.replace(range, suggestion);
            });
            // Strangely, the cursor moves during the edit, but the selection anchor
            // does not. Therefore, move the anchor to the cursor:
            textEditor.selection =
                new vscode_1.Selection(textEditor.selection.active, textEditor.selection.active);
        });
    }
    provideCodeActions(document, range, context) {
        const cmds = [];
        // filter diagnostic messages
        for (const diag of context.diagnostics) {
            if (diag.severity !== vscode_1.DiagnosticSeverity.Information) {
                continue;
            }
            // identify message
            const msg = this.server.messages.find((m) => m.file_name === document.fileName &&
                m.text === diag.message);
            // each "Try this" becomes a code action
            for (const [, tactic] of diag.message.matchAll(this.regexGM)) {
                cmds.push({
                    title: tactic,
                    command: '_lean.pasteTacticSuggestion',
                    arguments: [msg, tactic]
                });
            }
        }
        return cmds;
    }
}
exports.TacticSuggestions = TacticSuggestions;
//# sourceMappingURL=tacticsuggestions.js.map